From 505989f92fd6d97b42cf50739109651fde69bbe5 Mon Sep 17 00:00:00 2001
From: Shugo Maeda <shugo@ruby-lang.org>
Date: Fri, 27 Aug 2010 00:47:21 +0900
Subject: [PATCH 22/51] dup nd_omod lazily.

---
 eval.c          |   10 ++++++++--
 node.h          |    3 ++-
 vm_insnhelper.c |    2 +-
 vm_insnhelper.h |   15 +++++++++------
 4 files changed, 20 insertions(+), 10 deletions(-)

diff --git a/eval.c b/eval.c
index 06c7a90..d371b04 100644
--- a/eval.c
+++ b/eval.c
@@ -981,8 +981,14 @@ rb_overlay_module(NODE *cref, VALUE klass, VALUE module)
 	cref->nd_omod = rb_hash_new();
 	rb_funcall(cref->nd_omod, rb_intern("compare_by_identity"), 0);
     }
-    else if (!NIL_P(c = rb_hash_lookup(cref->nd_omod, klass))) {
-	superclass = c;
+    else {
+	if (cref->flags & NODE_FL_CREF_OMOD_SHARED) {
+	    cref->nd_omod = rb_hash_dup(cref->nd_omod);
+	    cref->flags &= ~NODE_FL_CREF_OMOD_SHARED;
+	}
+	if (!NIL_P(c = rb_hash_lookup(cref->nd_omod, klass))) {
+	    superclass = c;
+	}
     }
     c = iclass = rb_include_class_new(module, superclass);
     module = RCLASS_SUPER(module);
diff --git a/node.h b/node.h
index c798d96..405a469 100644
--- a/node.h
+++ b/node.h
@@ -267,8 +267,9 @@ typedef struct RNode {
 
 /* 0..4:T_TYPES, 5:FL_MARK, 6:NODE_FL_NESTED_DEF, 7:NODE_FL_NEWLINE */
 #define NODE_FL_NEWLINE (((VALUE)1)<<7)
-#define NODE_FL_CREF_PUSHED_BY_EVAL NODE_FL_NEWLINE
 #define NODE_FL_NESTED_DEF (((VALUE)1)<<6)
+#define NODE_FL_CREF_PUSHED_BY_EVAL NODE_FL_NEWLINE
+#define NODE_FL_CREF_OMOD_SHARED NODE_FL_NESTED_DEF
 
 #define NODE_TYPESHIFT 8
 #define NODE_TYPEMASK  (((VALUE)0x7f)<<NODE_TYPESHIFT)
diff --git a/vm_insnhelper.c b/vm_insnhelper.c
index 7c28085..edeefb8 100644
--- a/vm_insnhelper.c
+++ b/vm_insnhelper.c
@@ -1113,7 +1113,7 @@ vm_cref_push(rb_thread_t *th, VALUE klass, int noex, rb_block_t *blockptr)
     /* TODO: why cref->nd_next is 1? */
     if (cref->nd_next && cref->nd_next != (void *) 1 &&
 	!NIL_P(cref->nd_next->nd_omod)) {
-	cref->nd_omod = rb_hash_dup(cref->nd_next->nd_omod);
+	COPY_CREF_OMOD(cref, cref->nd_next);
     }
 
     return cref;
diff --git a/vm_insnhelper.h b/vm_insnhelper.h
index 3852ee5..ad3f129 100644
--- a/vm_insnhelper.h
+++ b/vm_insnhelper.h
@@ -149,14 +149,17 @@ extern VALUE ruby_vm_const_missing_count;
 /* deal with control flow 2: method/iterator              */
 /**********************************************************/
 
+#define COPY_CREF_OMOD(c1, c2) do {  \
+  c1->nd_omod = c2->nd_omod; \
+  if (!NIL_P(c2->nd_omod)) { \
+      c1->flags |= NODE_FL_CREF_OMOD_SHARED; \
+      c2->flags |= NODE_FL_CREF_OMOD_SHARED; \
+  } \
+} while (0)
+
 #define COPY_CREF(c1, c2) do {  \
   NODE *__tmp_c2 = (c2); \
-  if (NIL_P(__tmp_c2->nd_omod)) { \
-      c1->nd_omod = __tmp_c2->nd_omod; \
-  } \
-  else { \
-      c1->nd_omod = rb_hash_dup(__tmp_c2->nd_omod); \
-  } \
+  COPY_CREF_OMOD(c1, __tmp_c2); \
   c1->nd_clss = __tmp_c2->nd_clss; \
   c1->nd_visi = __tmp_c2->nd_visi;\
   c1->nd_next = __tmp_c2->nd_next; \
-- 
1.7.0.4

